# Do not modify this file by hand.
#
# It is automatically generated by src/api_wrappers/generatePythonAPIWrappers.py.
# (Run make api_wrappers to update it.)

from __future__ import print_function, unicode_literals, division, absolute_import

import sys

from dxpy import DXHTTPRequest
from dxpy.utils import Nonce

def analysis_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def analysis_terminate(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /analysis-xxxx/terminate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-analysis-xxxx-terminate
    """
    return DXHTTPRequest('/%s/terminate' % object_id, input_params, always_retry=always_retry, **kwargs)

def app_add_authorized_users(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-addauthorizedusers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_add_categories(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-addcategories
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_add_developers(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-adddevelopers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_add_tags(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-addtags
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_delete(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/delete API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-delete
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/delete' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_describe(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-describe
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/describe' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_get(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/get API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-get
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/get' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_install(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/install API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-install
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/install' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_list_authorized_users(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/listAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-listauthorizedusers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_list_categories(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/listCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-listcategories
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_list_developers(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/listDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-listdevelopers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_publish(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/publish API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-publish
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/publish' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_authorized_users(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removeauthorizedusers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_categories(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removecategories
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_developers(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removedevelopers
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_remove_tags(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-removetags
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_run(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/run' % fully_qualified_version, input_params_cp, always_retry=always_retry, **kwargs)

def app_validate_batch(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/validateBatch API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-validatebatch
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/validateBatch' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_uninstall(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/uninstall API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-uninstall
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/uninstall' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_update(app_name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-xxxx-yyyy-update
    """
    fully_qualified_version = app_name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/update' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def app_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /app/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/apps#api-method-app-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/app/new', input_params_cp, always_retry=always_retry, **kwargs)

def applet_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_get(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/get API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-get
    """
    return DXHTTPRequest('/%s/get' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_validate_batch(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/validateBatch API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-validatebatch
    """
    return DXHTTPRequest('/%s/validateBatch' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_run(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-xxxx-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/%s/run' % object_id, input_params_cp, always_retry=always_retry, **kwargs)

def applet_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def applet_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /applet/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-applet-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/applet/new', input_params_cp, always_retry=always_retry, **kwargs)

def container_clone(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/clone API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-clone
    """
    return DXHTTPRequest('/%s/clone' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/containers-for-execution#api-method-container-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_destroy(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/destroy API method.
    """
    return DXHTTPRequest('/%s/destroy' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_list_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/listFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-listfolder
    """
    return DXHTTPRequest('/%s/listFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_move(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/move API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-move
    """
    return DXHTTPRequest('/%s/move' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_new_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /container-xxxx/newFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-newfolder
    """
    return DXHTTPRequest('/%s/newFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_remove_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/removeFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removefolder
    """
    return DXHTTPRequest('/%s/removeFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_remove_objects(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/removeObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removeobjects
    """
    return DXHTTPRequest('/%s/removeObjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def container_rename_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /container-xxxx/renameFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-renamefolder
    """
    return DXHTTPRequest('/%s/renameFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle#api-method-class-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/databases#api-method-database-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_relocate(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /database-xxxx/relocate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/databases#api-method-database-xxxx-relocate
    """
    return DXHTTPRequest('/%s/relocate' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_download_file(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/downloadFile API method.
    """
    return DXHTTPRequest('/%s/downloadFile' % object_id, input_params, always_retry=always_retry, **kwargs)

def database_list_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /database-xxxx/listFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-listfolder
    """
    return DXHTTPRequest('/%s/listFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_new(input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /dbcluster/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-new
    """
    return DXHTTPRequest('/dbcluster/new', input_params, always_retry=always_retry, **kwargs)

def dbcluster_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_start(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/start API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-start
    """
    return DXHTTPRequest('/%s/start' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_stop(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/stop API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-stop
    """
    return DXHTTPRequest('/%s/stop' % object_id, input_params, always_retry=always_retry, **kwargs)

def dbcluster_terminate(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /dbcluster-xxxx/terminate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/dbclusters#api-method-dbcluster-xxxx-terminate
    """
    return DXHTTPRequest('/%s/terminate' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_download(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/download API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-download
    """
    return DXHTTPRequest('/%s/download' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_upload(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file-xxxx/upload API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-xxxx-upload
    """
    return DXHTTPRequest('/%s/upload' % object_id, input_params, always_retry=always_retry, **kwargs)

def file_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /file/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/files#api-method-file-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/file/new', input_params_cp, always_retry=always_retry, **kwargs)

def global_workflow_add_authorized_users(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-addauthorizedusers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_add_categories(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-addcategories
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_add_developers(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-adddevelopers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_add_tags(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-addtags
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/addTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_delete(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/delete API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-delete
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/delete' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_describe(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-describe
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/describe' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_list_authorized_users(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/listAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-listauthorizedusers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_list_categories(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/listCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-listcategories
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_list_developers(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/listDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-listdevelopers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/listDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_publish(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/publish API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-publish
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/publish' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_authorized_users(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeAuthorizedUsers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removeauthorizedusers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeAuthorizedUsers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_categories(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeCategories API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removecategories
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeCategories' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_developers(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeDevelopers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removedevelopers
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeDevelopers' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_remove_tags(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-removetags
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/removeTags' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_run(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/run' % fully_qualified_version, input_params_cp, always_retry=always_retry, **kwargs)

def global_workflow_update(name_or_id, alias=None, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-xxxx-yyyy-update
    """
    fully_qualified_version = name_or_id + (('/' + alias) if alias else '')
    return DXHTTPRequest('/%s/update' % fully_qualified_version, input_params, always_retry=always_retry, **kwargs)

def global_workflow_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /globalworkflow/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/global-workflows#api-method-globalworkflow-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/globalworkflow/new', input_params_cp, always_retry=always_retry, **kwargs)

def job_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_get_log(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /job-xxxx/getLog API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-getlog
    """
    return DXHTTPRequest('/%s/getLog' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_terminate(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/terminate API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-terminate
    """
    return DXHTTPRequest('/%s/terminate' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_get_identity_token(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job-xxxx/getIdentityToken API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-xxxx-getIdentityToken
    """
    return DXHTTPRequest('/%s/getIdentityToken' % object_id, input_params, always_retry=always_retry, **kwargs)

def job_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /job/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/applets-and-entry-points#api-method-job-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/job/new', input_params_cp, always_retry=always_retry, **kwargs)

def notifications_get(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /notifications/get API method.
    """
    return DXHTTPRequest('/notifications/get', input_params, always_retry=always_retry, **kwargs)

def notifications_mark_read(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /notifications/markRead API method.
    """
    return DXHTTPRequest('/notifications/markRead', input_params, always_retry=always_retry, **kwargs)

def org_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_find_members(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/findMembers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-findmembers
    """
    return DXHTTPRequest('/%s/findMembers' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_find_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/findProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-findprojects
    """
    return DXHTTPRequest('/%s/findProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_find_apps(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/findApps API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-findapps
    """
    return DXHTTPRequest('/%s/findApps' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_invite(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/invite API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-invite
    """
    return DXHTTPRequest('/%s/invite' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_remove_member(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/removeMember API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-removemember
    """
    return DXHTTPRequest('/%s/removeMember' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_set_member_access(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/setMemberAccess API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-setmemberaccess
    """
    return DXHTTPRequest('/%s/setMemberAccess' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def org_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /org/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/organizations#api-method-org-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/org/new', input_params_cp, always_retry=always_retry, **kwargs)

def project_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_archive(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/archive API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-archive
    """
    return DXHTTPRequest('/%s/archive' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_unarchive(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/unarchive API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-unarchive
    """
    return DXHTTPRequest('/%s/unarchive' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_clone(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/clone API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-clone
    """
    return DXHTTPRequest('/%s/clone' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_decrease_permissions(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/decreasePermissions API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-decreasepermissions
    """
    return DXHTTPRequest('/%s/decreasePermissions' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_destroy(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/destroy API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-destroy
    """
    return DXHTTPRequest('/%s/destroy' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_invite(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/invite API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-invite
    """
    return DXHTTPRequest('/%s/invite' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_leave(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/leave API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-leave
    """
    return DXHTTPRequest('/%s/leave' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_list_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/listFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-listfolder
    """
    return DXHTTPRequest('/%s/listFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_move(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/move API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-move
    """
    return DXHTTPRequest('/%s/move' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_new_folder(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/newFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-newfolder
    """
    return DXHTTPRequest('/%s/newFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_remove_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/removeFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removefolder
    """
    return DXHTTPRequest('/%s/removeFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_remove_objects(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/removeObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-removeobjects
    """
    return DXHTTPRequest('/%s/removeObjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_rename_folder(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project-xxxx/renameFolder API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/folders-and-deletion#api-method-class-xxxx-renamefolder
    """
    return DXHTTPRequest('/%s/renameFolder' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_transfer(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/transfer API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/project-permissions-and-sharing#api-method-project-xxxx-transfer
    """
    return DXHTTPRequest('/%s/transfer' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_update_sponsorship(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /project-xxxx/updateSponsorship API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-xxxx-updatesponsorship
    """
    return DXHTTPRequest('/%s/updateSponsorship' % object_id, input_params, always_retry=always_retry, **kwargs)

def project_new(input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /project/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/projects#api-method-project-new
    """
    return DXHTTPRequest('/project/new', input_params, always_retry=always_retry, **kwargs)

def record_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle#api-method-class-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/records#api-method-record-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def record_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /record/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-classes/records#api-method-record-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/record/new', input_params_cp, always_retry=always_retry, **kwargs)

def system_describe_data_objects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/describeDataObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-describedataobjects
    """
    return DXHTTPRequest('/system/describeDataObjects', input_params, always_retry=always_retry, **kwargs)

def system_describe_executions(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/describeExecutions API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-describeexecutions
    """
    return DXHTTPRequest('/system/describeExecutions', input_params, always_retry=always_retry, **kwargs)

def system_describe_projects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/describeProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-describeprojects
    """
    return DXHTTPRequest('/system/describeProjects', input_params, always_retry=always_retry, **kwargs)

def system_find_affiliates(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findAffiliates API method.
    """
    return DXHTTPRequest('/system/findAffiliates', input_params, always_retry=always_retry, **kwargs)

def system_find_apps(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findApps API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findapps
    """
    return DXHTTPRequest('/system/findApps', input_params, always_retry=always_retry, **kwargs)

def system_find_data_objects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findDataObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-finddataobjects
    """
    return DXHTTPRequest('/system/findDataObjects', input_params, always_retry=always_retry, **kwargs)

def system_find_global_workflows(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findGlobalWorkflows API method.
    """
    return DXHTTPRequest('/system/findGlobalWorkflows', input_params, always_retry=always_retry, **kwargs)

def system_resolve_data_objects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/resolveDataObjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-resolvedataobjects
    """
    return DXHTTPRequest('/system/resolveDataObjects', input_params, always_retry=always_retry, **kwargs)

def system_find_executions(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findExecutions API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findexecutions
    """
    return DXHTTPRequest('/system/findExecutions', input_params, always_retry=always_retry, **kwargs)

def system_find_analyses(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findAnalyses API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findanalyses
    """
    return DXHTTPRequest('/system/findAnalyses', input_params, always_retry=always_retry, **kwargs)

def system_find_databases(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findDatabases API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-finddatabases
    """
    return DXHTTPRequest('/system/findDatabases', input_params, always_retry=always_retry, **kwargs)

def system_find_jobs(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findJobs API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findjobs
    """
    return DXHTTPRequest('/system/findJobs', input_params, always_retry=always_retry, **kwargs)

def system_find_projects(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findprojects
    """
    return DXHTTPRequest('/system/findProjects', input_params, always_retry=always_retry, **kwargs)

def system_find_users(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findUsers API method.
    """
    return DXHTTPRequest('/system/findUsers', input_params, always_retry=always_retry, **kwargs)

def system_find_project_members(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findProjectMembers API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findprojectmembers
    """
    return DXHTTPRequest('/system/findProjectMembers', input_params, always_retry=always_retry, **kwargs)

def system_find_orgs(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/findOrgs API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/search#api-method-system-findorgs
    """
    return DXHTTPRequest('/system/findOrgs', input_params, always_retry=always_retry, **kwargs)

def system_generate_batch_inputs(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/generateBatchInputs API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-generatebatchinputs
    """
    return DXHTTPRequest('/system/generateBatchInputs', input_params, always_retry=always_retry, **kwargs)

def system_global_search(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/globalSearch API method.
    """
    return DXHTTPRequest('/system/globalSearch', input_params, always_retry=always_retry, **kwargs)

def system_greet(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/greet API method.
    """
    return DXHTTPRequest('/system/greet', input_params, always_retry=always_retry, **kwargs)

def system_headers(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/headers API method.
    """
    return DXHTTPRequest('/system/headers', input_params, always_retry=always_retry, **kwargs)

def system_shorten_url(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/shortenURL API method.
    """
    return DXHTTPRequest('/system/shortenURL', input_params, always_retry=always_retry, **kwargs)

def system_whoami(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /system/whoami API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/system-methods#api-method-system-whoamiwiki.
    """
    return DXHTTPRequest('/system/whoami', input_params, always_retry=always_retry, **kwargs)

def user_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /user-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/users#api-method-user-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def user_update(object_id, input_params={}, always_retry=False, **kwargs):
    """
    Invokes the /user-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/users#api-method-user-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_add_stage(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/addStage API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-addstage
    """
    return DXHTTPRequest('/%s/addStage' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_add_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/addTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-addtags
    """
    return DXHTTPRequest('/%s/addTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_add_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/addTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-addtypes
    """
    return DXHTTPRequest('/%s/addTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_close(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/close API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle#api-method-class-xxxx-close
    """
    return DXHTTPRequest('/%s/close' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_describe(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/describe API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-describe
    """
    return DXHTTPRequest('/%s/describe' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_dry_run(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/dryRun API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-dryrun
    """
    return DXHTTPRequest('/%s/dryRun' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_get_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/getDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-getdetails
    """
    return DXHTTPRequest('/%s/getDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_is_stage_compatible(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/isStageCompatible API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-isstagecompatible
    """
    return DXHTTPRequest('/%s/isStageCompatible' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_list_projects(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/listProjects API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-containers/cloning#api-method-class-xxxx-listprojects
    """
    return DXHTTPRequest('/%s/listProjects' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_move_stage(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/moveStage API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-movestage
    """
    return DXHTTPRequest('/%s/moveStage' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_overwrite(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/overwrite API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-overwrite
    """
    return DXHTTPRequest('/%s/overwrite' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_remove_stage(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/removeStage API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-removestage
    """
    return DXHTTPRequest('/%s/removeStage' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_remove_tags(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/removeTags API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/tags#api-method-class-xxxx-removetags
    """
    return DXHTTPRequest('/%s/removeTags' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_remove_types(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/removeTypes API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/types#api-method-class-xxxx-removetypes
    """
    return DXHTTPRequest('/%s/removeTypes' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_rename(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/rename API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/name#api-method-class-xxxx-rename
    """
    return DXHTTPRequest('/%s/rename' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_run(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/run API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-run
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/%s/run' % object_id, input_params_cp, always_retry=always_retry, **kwargs)

def workflow_validate_batch(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/validateBatch API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-validatebatch
    """
    return DXHTTPRequest('/%s/validateBatch' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_set_details(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/setDetails API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/details-and-links#api-method-class-xxxx-setdetails
    """
    return DXHTTPRequest('/%s/setDetails' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_set_properties(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/setProperties API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/introduction-to-data-object-metadata/properties#api-method-class-xxxx-setproperties
    """
    return DXHTTPRequest('/%s/setProperties' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_set_visibility(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/setVisibility API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/data-object-lifecycle/visibility#api-method-class-xxxx-setvisibility
    """
    return DXHTTPRequest('/%s/setVisibility' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_update(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/update API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-update
    """
    return DXHTTPRequest('/%s/update' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_update_stage_executable(object_id, input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow-xxxx/updateStageExecutable API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-xxxx-updatestageexecutable
    """
    return DXHTTPRequest('/%s/updateStageExecutable' % object_id, input_params, always_retry=always_retry, **kwargs)

def workflow_new(input_params={}, always_retry=True, **kwargs):
    """
    Invokes the /workflow/new API method.

    For more info, see: https://documentation.dnanexus.com/developer/api/running-analyses/workflows-and-analyses#api-method-workflow-new
    """
    input_params_cp = Nonce.update_nonce(input_params)
    return DXHTTPRequest('/workflow/new', input_params_cp, always_retry=always_retry, **kwargs)




# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""
Utilities shared by dxpy modules.
"""

from __future__ import print_function, unicode_literals, division, absolute_import

import os, json, collections, concurrent.futures, traceback, sys, time, gc, platform
from multiprocessing import cpu_count
from collections.abc import Mapping
import dateutil.parser
from .. import logger
from ..compat import basestring, THREAD_TIMEOUT_MAX
from ..exceptions import DXError
import numbers
import binascii
import random

def _force_quit(signum, frame):
    # traceback.print_stack(frame)
    os._exit(os.EX_IOERR)

def get_futures_threadpool(max_workers):
    return concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)

def wait_for_a_future(futures, print_traceback=False):
    """
    Return the next future that completes.  If a KeyboardInterrupt is
    received, then the entire process is exited immediately.  See
    wait_for_all_futures for more notes.
    """
    while True:
        try:
            future = next(concurrent.futures.as_completed(futures, timeout=THREAD_TIMEOUT_MAX))
            break
        except concurrent.futures.TimeoutError:
            pass
        except KeyboardInterrupt:
            if print_traceback:
                traceback.print_stack()
            else:
                print('')
            os._exit(os.EX_IOERR)

    return future

def wait_for_all_futures(futures, print_traceback=False):
    """
    Wait indefinitely for all futures in the input iterable to complete.
    Use a timeout to enable interrupt handling.
    Call os._exit() in case of KeyboardInterrupt. Otherwise, the atexit registered handler in concurrent.futures.thread
    will run, and issue blocking join() on all worker threads, requiring us to listen to events in worker threads
    in order to enable timely exit in response to Ctrl-C.

    Note: This still doesn't handle situations where Ctrl-C is pressed elsewhere in the code and there are worker
    threads with long-running tasks.

    Note: os._exit() doesn't work well with interactive mode (e.g. ipython). This may help:
    import __main__ as main; if hasattr(main, '__file__'): os._exit() else: os.exit()
    """
    try:
        while True:
            waited_futures = concurrent.futures.wait(futures, timeout=60)
            if len(waited_futures.not_done) == 0:
                break
    except KeyboardInterrupt:
        if print_traceback:
            traceback.print_stack()
        else:
            print('')
        os._exit(os.EX_IOERR)


def response_iterator(request_iterator, thread_pool, max_active_tasks=None, do_first_task_sequentially=True):
    """
    :param request_iterator:
        An iterator producing inputs for consumption by the worker pool.
    :type request_iterator: iterator of callable, args, kwargs
    :param thread_pool: thread pool to submit the requests to
    :type thread_pool: concurrent.futures.thread.ThreadPoolExecutor
    :param max_active_tasks:
        The maximum number of tasks that may be either running or
        waiting for consumption of their result. If not given, defaults
        to the number of CPU cores on the machine.
    :type max_active_tasks: int
    :param do_first_task_sequentially:
        If True, executes (and returns the result of) the first request
        before submitting any other requests (the subsequent requests
        are submitted with *max_active_tasks* parallelism).
    :type do_first_task_sequentially: bool

    Rate-limited asynchronous multithreaded task runner. Consumes tasks
    from *request_iterator*. Yields their results in order, while
    allowing up to *max_active_tasks* to run simultaneously. Unlike
    concurrent.futures.Executor.map, prevents new tasks from starting
    while there are *max_active_tasks* or more unconsumed results.

    """
    tasks_in_progress = collections.deque()
    if max_active_tasks is None:
        max_active_tasks = cpu_count()

    # The following two functions facilitate GC by not adding extra variables to the enclosing scope.
    def submit_task(task_iterator, executor, futures_queue):
        retval  = next(task_iterator, None)
        if retval is None:
            return False
        task_callable, task_args, task_kwargs = retval
        task_future = executor.submit(task_callable, *task_args, **task_kwargs)
        futures_queue.append(task_future)
        return True

    def next_result(tasks_in_progress):
        future = tasks_in_progress.popleft()
        try:
            result = future.result(timeout=THREAD_TIMEOUT_MAX)
        except KeyboardInterrupt:
            print('')
            os._exit(os.EX_IOERR)
        return result

    if do_first_task_sequentially:
        task_callable, task_args, task_kwargs = next(request_iterator)
        yield task_callable(*task_args, **task_kwargs)

    for _i in range(max_active_tasks):
        retval = submit_task(request_iterator, thread_pool, tasks_in_progress)
        if not retval:
            break

    while len(tasks_in_progress) > 0:
        result = next_result(tasks_in_progress)
        submit_task(request_iterator, thread_pool, tasks_in_progress)
        yield result
        del result

def string_buffer_length(buf):
    orig_pos = buf.tell()
    buf.seek(0, os.SEEK_END)
    buf_len = buf.tell()
    buf.seek(orig_pos)
    return buf_len


def normalize_time_input(t, future=False, default_unit='ms'):
    """
    :param default_unit: units of the input time *t*; must be one of "s" or
        "ms". This param is only respected if *t* looks like an int (e.g.
        "12345", 12345).
    :type default_unit: string

    Converts inputs such as:
       "2012-05-01"
       "-5d"
       1352863174
       "1352863174"
    to milliseconds since epoch. See http://labix.org/python-dateutil and :meth:`normalize_timedelta`.
    """
    error_msg = 'Error: Expected an int timestamp, a date format (e.g. YYYY-MM-DD), or an int with a single-letter suffix (s=seconds, m=minutes, h=hours, d=days, w=weeks, M=months, y=years; e.g. "-10d" indicates 10 days ago); but got {t}'
    if isinstance(t, basestring) and t.isdigit():
        t = int(t)

    if isinstance(t, basestring):
        try:
            t = normalize_timedelta(t)
        except ValueError:
            try:
                t = int(time.mktime(dateutil.parser.parse(t).timetuple())*1000)
                assert t > 0
            except (ValueError, OverflowError, AssertionError):
                raise ValueError(error_msg.format(t=t))
    elif isinstance(t, numbers.Integral):
        units_multipliers = {'ms': 1, 's': 1000}
        if default_unit not in units_multipliers:
            raise ValueError("Expected default_unit to be one of 's' or 'ms'")
        t = t * units_multipliers[default_unit]
    else:
        raise ValueError(error_msg.format(t=t))
    now = int(time.time()*1000)
    if t < 0 or (future and t < now):
        t += now
    return t

def normalize_timedelta(timedelta):
    """
    Given a string like "1w" or "-5d", convert it to an integer in milliseconds.
    Integers without a suffix are interpreted as seconds.
    Note: not related to the datetime timedelta class.
    """
    try:
        return int(timedelta) * 1000
    except ValueError as e:
        t, suffix = timedelta[:-1], timedelta[-1:]
        suffix_multipliers = {'s': 1000, 'm': 1000*60, 'h': 1000*60*60, 'd': 1000*60*60*24, 'w': 1000*60*60*24*7,
                              'M': 1000*60*60*24*30, 'y': 1000*60*60*24*365}
        if suffix not in suffix_multipliers:
            raise ValueError()
        return int(t) * suffix_multipliers[suffix]

# See http://stackoverflow.com/questions/4126348
class OrderedDefaultdict(collections.OrderedDict):
    def __init__(self, *args, **kwargs):
        newdefault = None
        newargs = ()
        if args:
            newdefault = args[0]
            if not (newdefault is None or callable(newdefault)):
                raise TypeError('first argument must be callable or None')
            newargs = args[1:]
        self.default_factory = newdefault
        super(self.__class__, self).__init__(*newargs, **kwargs)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        args = self.default_factory if self.default_factory else tuple()
        return type(self), args, None, None, self.items()

def group_array_by_field(array, field='group'):
    groups = OrderedDefaultdict(list)
    for item in array:
        if field not in item and None not in groups:
            groups[None] = []
        groups[item.get(field)].append(item)
    return groups

def merge(d, u):
    """
    Recursively updates a dictionary.
    Example: merge({"a": {"b": 1, "c": 2}}, {"a": {"b": 3}}) = {"a": {"b": 3, "c": 2}}
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            r = merge(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d

def _dict_raise_on_duplicates(ordered_pairs):
    """
    Reject duplicate keys.
    """
    d = {}
    for k, v in ordered_pairs:
        if k in d:
           raise ValueError("duplicate key: %r" % (k,))
        else:
           d[k] = v
    return d

def json_load_raise_on_duplicates(*args, **kwargs):
    """
    Like json.load(), but raises an error on duplicate keys.
    """
    kwargs['object_pairs_hook'] = _dict_raise_on_duplicates
    return json.load(*args, **kwargs)

def json_loads_raise_on_duplicates(*args, **kwargs):
    """
    Like json.loads(), but raises an error on duplicate keys.
    """
    kwargs['object_pairs_hook'] = _dict_raise_on_duplicates
    return json.loads(*args, **kwargs)

def warn(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

class Nonce:
    '''
    Generates a nonce by using the system's random number generator. If it fails
    it uses python's random library to generate a random long integer.
    The nonce is the random number concatenated with the time.
    '''
    def __init__(self):
        try:
            self.nonce = "%s%f" % (binascii.hexlify(os.urandom(32)).decode('utf-8'), time.time())
        except:
            random.seed(time.time())
            self.nonce = "%s%f" % (random.getrandbits(8*26), time.time())

    def __str__(self):
        return self.nonce

    @staticmethod
    def update_nonce(input_params):
        '''
        Static method to return a copy of the input dictionary with an
        additional unique nonce
        :param input: an input dictionary that may be empty
        :type input: dict
        :returns an extended copy of the input with an additional nonce field

        The input dictionary is updated with a nonce only if does not already
        have a non empty nonce
        '''
        input_cp = input_params.copy()
        if len(input_cp.get('nonce', '')) == 0:
            input_cp['nonce'] = str(Nonce())
        return input_cp

# Moved to the bottom due to circular imports
from .exec_utils import run, convert_handlers_to_dxlinks, parse_args_as_job_input, entry_point, DXJSONEncoder



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import json
import os.path
import subprocess
from dxpy.nextflow.ImageRefFactory import ImageRefFactory, ImageRefFactoryError

CONTAINERS_JSON = "containers.json"


def bundle_docker_images(image_refs):
    """
    :param image_refs: Image references extracted from run_nextaur_collect().
    :type image_refs: Dict
    :returns: Array of dicts for bundledDepends attribute of the applet resources. Also saves images on the platform
    if not done that before.
    """
    image_factories = [ImageRefFactory(x) for x in image_refs]
    images = [x.get_image() for x in image_factories]
    seen_images = set()
    bundled_depends = []
    for image in images:
        if image.identifier in seen_images:
            continue
        else:
            bundled_depends.append(image.bundled_depends.copy())
            seen_images.add(image.identifier)
    return bundled_depends


def run_nextaur_collect(resources_dir, profile, nextflow_pipeline_params):
    """
        :param resources_dir: URL to the local(ized) NF pipeline in the app(let) resources.
        :type resources_dir: String
        :param profile: Custom Nextflow profile. More profiles can be provided by using comma separated string (without whitespaces).
        :type profile: str
        :param nextflow_pipeline_params: Custom Nextflow pipeline parameters
        :type nextflow_pipeline_params: string
        :returns: Dict. Image references in the form of
            "process": String. Name of the process/task
            "repository": String. Repository (host) prefix
            "image_name": String. Image base name
            "tag": String. Version tag
            "digest": String. Image digest
            "file_id": String. File ID if found on the platform
            "engine": String. Container engine.
        Runs nextaur:collect
        """
    base_cmd = "nextflow plugin nextaur:collect docker {}".format(resources_dir)
    pipeline_params_arg = "pipelineParams={}".format(nextflow_pipeline_params) if nextflow_pipeline_params else ""
    profile_arg = "profile={}".format(profile) if profile else ""
    nextaur_cmd = " ".join([base_cmd, pipeline_params_arg, profile_arg])
    process = subprocess.run(nextaur_cmd, shell=True, capture_output=True, text=True)
    if not os.path.exists(CONTAINERS_JSON):
        raise ImageRefFactoryError(process.stdout)
    with open(CONTAINERS_JSON, "r") as json_file:
        image_refs = json.load(json_file).get("processes", None)
        if not image_refs:
            raise ImageRefFactoryError("Could not extract processes from nextaur:collect")
    return image_refs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import print_function, unicode_literals, division, absolute_import

import os
import unittest
import tempfile
import json
import sys
import shutil
import subprocess
import pytest

import dxpy
import dxpy_testutil as testutil
from dxpy_testutil import (DXTestCase, check_output, override_environment, chdir)


def run(command, **kwargs):
    print("$ %s" % (command,))
    output = check_output(command, shell=True, **kwargs)
    print(output)
    return output


class TestDXBuildAsset(DXTestCase):
    def setUp(self):
        super(TestDXBuildAsset, self).setUp()
        self.temp_file_path = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_file_path)
        super(TestDXBuildAsset, self).tearDown()

    def write_asset_directory(self, asset_name, dxasset_str, asset_rsc_dir=None, makefile_str=None):
        # Note: if called twice with the same asset_name, will overwrite
        # the dxasset.json
        try:
            os.mkdir(os.path.join(self.temp_file_path, asset_name))
        except OSError as e:
            if e.errno != 17:
                raise e
        if dxasset_str is not None:
            with open(os.path.join(self.temp_file_path, asset_name, 'dxasset.json'), 'wb') as manifest:
                manifest.write(dxasset_str.encode())
        if asset_rsc_dir:
            try:
                os.mkdir(os.path.join(self.temp_file_path, asset_name, asset_rsc_dir))
            except OSError as e:
                if e.errno != 17:
                    raise e
        if makefile_str is not None:
            with open(os.path.join(self.temp_file_path, asset_name, 'Makefile'), 'wb') as manifest:
                manifest.write(makefile_str.encode())

        return os.path.join(self.temp_file_path, asset_name)

    def write_app_directory(self, app_name, dxapp_str, code_filename=None, code_content="\n"):
        # Note: if called twice with the same app_name, will overwrite
        # the dxapp.json and code file (if specified) but will not
        # remove any other files that happened to be present
        try:
            os.mkdir(os.path.join(self.temp_file_path, app_name))
        except OSError as e:
            if e.errno != 17:  # directory already exists
                raise e
        if dxapp_str is not None:
            with open(os.path.join(self.temp_file_path, app_name, 'dxapp.json'), 'wb') as manifest:
                manifest.write(dxapp_str.encode())
        if code_filename:
            with open(os.path.join(self.temp_file_path, app_name, code_filename), 'w') as code_file:
                code_file.write(code_content)
        return os.path.join(self.temp_file_path, app_name)

    def test_build_asset_help(self):
        env = override_environment(DX_SECURITY_CONTEXT=None, DX_APISERVER_HOST=None,
                                   DX_APISERVER_PORT=None, DX_APISERVER_PROTOCOL=None)
        run("dx build_asset -h", env=env)

    def test_build_asset_with_no_dxasset_json(self):
        asset_dir = self.write_asset_directory("asset_with_no_json", None)
        with self.assertSubprocessFailure(stderr_regexp='is not a valid DNAnexus asset source directory',
                                          exit_code=1):
            run("dx build_asset " + asset_dir)

    def test_build_asset_with_malformed_dxasset_json(self):
        asset_dir = self.write_asset_directory("asset_with_malform_json", "{")
        with self.assertSubprocessFailure(stderr_regexp=r'Could not parse dxasset\.json', exit_code=1):
            run("dx build_asset " + asset_dir)

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    @pytest.mark.TRACEABILITY_MATRIX
    @testutil.update_traceability_matrix(["DNA_CLI_DATA_OBJ_BUILD_ASSET"])
    def test_build_asset_with_valid_dxasset(self):
        asset_spec = {
            "name": "asset_library_name",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "20.04",
            "version": "0",
            "instanceType": "mem1_ssd1_x2",
            "execDepends": [{"name": "python-numpy"}]
        }
        asset_dir = self.write_asset_directory("asset_with_valid_json", json.dumps(asset_spec))
        asset_bundle_id = json.loads(run('dx build_asset --json ' + asset_dir))['id']
        self.assertIn('record', asset_bundle_id)
        self.assertEqual(dxpy.describe(asset_bundle_id)['project'], self.project)
        job_id = dxpy.describe(asset_bundle_id)['createdBy']['job']
        self.assertEqual(dxpy.describe(job_id)['instanceType'], "mem1_ssd1_x2")

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_build_asset_with_valid_destination(self):
        asset_spec = {
            "name": "foo",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04",
            "execDepends": [{"name": "python-numpy"}]
        }
        asset_dir = self.write_asset_directory("asset_with_valid_destination", json.dumps(asset_spec))
        with testutil.temporary_project() as other_project:
            test_dirname = 'asset_dir'
            run('dx mkdir -p {project}:{dirname}'.format(project=other_project.get_id(), dirname=test_dirname))
            asset_bundle_id = json.loads(run('dx build_asset --json --destination ' + other_project.get_id() +
                                             ':/' + test_dirname + '/ ' + asset_dir))['id']
            self.assertIn('record', asset_bundle_id)
            asset_desc = dxpy.describe(asset_bundle_id)
            self.assertEqual(asset_desc['project'], other_project.get_id())
            self.assertEqual(asset_desc['folder'], '/asset_dir')

    def test_build_asset_invalid_destination(self):
        asset_spec = {
            "name": "asset_library_name",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("asset_with_invalid_destination", json.dumps(asset_spec))
        with self.assertSubprocessFailure(stderr_regexp='Could not find a project named', exit_code=3):
            run("dx build_asset -d nonexistent_project:/new-name " + asset_dir)

    def test_build_asset_missing_fields(self):
        asset_spec = {
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("asset_with_missing_fields", json.dumps(asset_spec))
        with self.assertSubprocessFailure(stderr_regexp='The asset configuration does not contain', exit_code=1):
            run("dx build_asset " + asset_dir)

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_build_asset_with_resources(self):
        asset_spec = {
            "name": "asset_library_name",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("asset_with_resources", json.dumps(asset_spec), "resources")
        asset_bundle_id = json.loads(run('dx build_asset --json ' + asset_dir))['id']
        self.assertIn('record', asset_bundle_id)
        self.assertEqual(dxpy.describe(asset_bundle_id)['project'], self.project)

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_build_asset_with_invalid_makefile(self):
        asset_spec = {
            "name": "asset_library_name",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("asset_with_invalid_makefile", json.dumps(asset_spec),
                                               None, "echo")
        # TODO exit_code=3
        with self.assertSubprocessFailure(stderr_regexp='', exit_code=1):
            run("dx build_asset --json " + asset_dir)

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_build_and_use_asset(self):
        asset_spec = {
            "name": "asset library name with space",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("build_and_use_asset", json.dumps(asset_spec), "resources")

        run("mkdir -p " + os.path.join(asset_dir, "resources/usr/local/bin"))
        with open(os.path.join(asset_dir, "resources/usr/local/bin", 'test.sh'), 'wb') as manifest:
            manifest.write("echo 'hi'".encode())
        run("chmod +x " + os.path.join(asset_dir, "resources/usr/local/bin", 'test.sh'))

        asset_bundle_id = json.loads(run('dx build_asset --json ' + asset_dir))['id']
        code_str = """#!/bin/bash
                    main(){
                        test.sh
                    }
                    """
        app_spec = {
            "name": "asset_depends",
            "dxapi": "1.0.0",
            "runSpec": {
                "code": code_str,
                "interpreter": "bash",
                "distribution": "Ubuntu",
                "release": "14.04",
                "assetDepends":  [{"id": asset_bundle_id}]
            },
            "inputSpec": [],
            "outputSpec": [],
            "version": "1.0.0"
        }
        app_dir = self.write_app_directory("asset_depends", json.dumps(app_spec))
        asset_applet_id = json.loads(run("dx build --json {app_dir}".format(app_dir=app_dir)))["id"]
        asset_applet = dxpy.DXApplet(asset_applet_id)
        applet_job = asset_applet.run({})
        applet_job.wait_on_done()
        self.assertEqual(applet_job.describe()['state'], 'done')

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_build_and_use_asset_with_stages(self):
        asset_spec = {
            "name": "asset library name with space",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("build_and_use_asset", json.dumps(asset_spec), "resources")

        run("mkdir -p " + os.path.join(asset_dir, "resources/usr/local/bin"))
        run("mkdir -p " + os.path.join(asset_dir, "resources/home/dnanexus"))
        with open(os.path.join(asset_dir, "resources/home/dnanexus", 'file_inside_asset.txt'), 'wb') as manifest:
            manifest.write("echo 'hi'".encode())
        with open(os.path.join(asset_dir, "resources/usr/local/bin", 'test.sh'), 'wb') as manifest:
            manifest.write("echo 'hi'".encode())
        run("chmod +x " + os.path.join(asset_dir, "resources/usr/local/bin", 'test.sh'))

        asset_bundle_id = json.loads(run('dx build_asset --json ' + asset_dir))['id']
        code_str = """#!/bin/bash
                    main(){
                        test.sh
                        dx-jobutil-new-job stage_without_asset
                    }
                    stage_without_asset(){
                        if [[ -e file_inside_asset.txt ]]; then
                            echo "file should not be here"
                            exit 1
                        fi
                    }
                    """
        app_spec = {
            "name": "asset_depends",
            "dxapi": "1.0.0",
            "runSpec": {
                "code": code_str,
                "interpreter": "bash",
                "distribution": "Ubuntu",
                "release": "14.04",
                "assetDepends":  [{"id": asset_bundle_id, "stages": ["main"]}]
            },
            "inputSpec": [],
            "outputSpec": [],
            "version": "1.0.0"
        }
        app_dir = self.write_app_directory("asset_depends", json.dumps(app_spec))
        asset_applet_id = json.loads(run("dx build --json {app_dir}".format(app_dir=app_dir)))["id"]
        asset_applet = dxpy.DXApplet(asset_applet_id)
        applet_job = asset_applet.run({})
        applet_job.wait_on_done()
        self.assertEqual(applet_job.describe()['state'], 'done')

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_build_asset_with_invalid_instance_type(self):
        asset_spec = {
            "name": "asset_library_name",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04",
            "instanceType": "mem1_ssd1_x3"
        }
        asset_dir = self.write_asset_directory("build_asset_with_invalid_instance_type", json.dumps(asset_spec))
        # TODO exit_code=3
        with self.assertSubprocessFailure(stderr_regexp='code 422', exit_code=1):
            run("dx build_asset --json " + asset_dir)

    @unittest.skipUnless(testutil.TEST_ONLY_MASTER and testutil.TEST_RUN_JOBS,
                         'skipping test that requires latest server version')
    def test_build_asset_inside_job(self):
        asset_spec = {
            "name": "asset library name with space",
            "title": "A human readable name",
            "description": " A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("test_build_asset_inside_job", json.dumps(asset_spec))
        asset_conf_file_id = run("dx upload " + os.path.join(asset_dir, "dxasset.json") + " --brief --wait").strip()
        code_str = """#!/bin/bash
                    main(){
                        dx download "${asset_conf}" -o dxasset.json
                        dx build_asset
                    }
                    """
        app_spec = {
            "name": "run_build_asset",
            "dxapi": "1.0.0",
            "runSpec": {
                "code": code_str,
                "interpreter": "bash",
                "distribution": "Ubuntu",
                "release": "14.04"
            },
            "inputSpec": [{"name": "asset_conf", "class": "file"}],
            "outputSpec": [],
            "version": "1.0.0"
        }
        app_dir = self.write_app_directory("run_build_asset", json.dumps(app_spec))
        asset_applet_id = json.loads(run("dx build --json {app_dir}".format(app_dir=app_dir)))["id"]

        asset_applet = dxpy.DXApplet(asset_applet_id)
        applet_job = asset_applet.run({"asset_conf": {"$dnanexus_link": asset_conf_file_id}})
        applet_job.wait_on_done()
        self.assertEqual(applet_job.describe()['state'], 'done')

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_get_appet_with_asset(self):
        bundle_name = "test-bundle-depends.tar.gz"
        asset_name = "test-asset-depends.tar.gz"
        bundle_tmp_dir = tempfile.mkdtemp()
        os.mkdir(os.path.join(bundle_tmp_dir, "a"))
        with open(os.path.join(bundle_tmp_dir, 'a', 'foo.txt'), 'w') as file_in_bundle:
            file_in_bundle.write('foo\n')
        subprocess.check_call(['tar', '-czf', os.path.join(bundle_tmp_dir, bundle_name),
                               '-C', os.path.join(bundle_tmp_dir, 'a'), '.'])
        subprocess.check_call(['tar', '-czf', os.path.join(bundle_tmp_dir, asset_name),
                               '-C', os.path.join(bundle_tmp_dir, 'a'), '.'])
        bundle_file = dxpy.upload_local_file(filename=os.path.join(bundle_tmp_dir, bundle_name),
                                             project=self.project,
                                             wait_on_close=True)

        asset_file = dxpy.upload_local_file(filename=os.path.join(bundle_tmp_dir, asset_name),
                                            project=self.project,
                                            wait_on_close=True)

        dxrecord_details = {"archiveFileId": {"$dnanexus_link": asset_file.get_id()}}
        dxrecord = dxpy.new_dxrecord(project=self.project, types=["AssetBundle"], details=dxrecord_details,
                                     name='asset-lib-test', properties={"version": "0.0.1"})
        dxrecord.close()
        asset_bundle_id = dxrecord.get_id()

        asset_file.set_properties({"AssetBundle": asset_bundle_id})

        code_str = """#!/bin/bash
                    main(){
                        echo 'Hello World'
                    }
                    """
        app_spec = {
            "name": "asset_depends",
            "dxapi": "1.0.0",
            "runSpec": {
                "code": code_str,
                "interpreter": "bash",
                "distribution": "Ubuntu",
                "release": "14.04",
                "assetDepends":  [{"id": asset_bundle_id}],
                "bundledDepends": [{"name": bundle_name, "id": {"$dnanexus_link": bundle_file.get_id()}}]
            },
            "inputSpec": [],
            "outputSpec": [],
            "version": "1.0.0"
        }
        app_dir = self.write_app_directory("asset_depends", json.dumps(app_spec))
        asset_applet_id = json.loads(run("dx build --json {app_dir}".format(app_dir=app_dir)))["id"]
        with chdir(tempfile.mkdtemp()):
            run("dx get --omit-resources " + asset_applet_id)
            self.assertTrue(os.path.exists("asset_depends"))
            self.assertFalse(os.path.exists(os.path.join("asset_depends", "resources")))
            self.assertTrue(os.path.exists(os.path.join("asset_depends", "dxapp.json")))
            with open(os.path.join("asset_depends", "dxapp.json")) as fh:
                applet_spec = json.load(fh)
            current_region = dxpy.describe(self.project).get("region")
            regional_options = applet_spec["regionalOptions"][current_region]
            self.assertIn({"name": bundle_name, "id": {"$dnanexus_link": bundle_file.get_id()}},
                             regional_options["bundledDepends"])
            self.assertIn({"name": asset_name, "id": {"$dnanexus_link": asset_file.get_id()}},
                             regional_options["bundledDepends"])

    @unittest.skipUnless(testutil.TEST_RUN_JOBS, 'skipping test that would run jobs')
    def test_set_assetbundle_tarball_property(self):
        asset_spec = {
            "name": "tarball_property_assetbundle",
            "title": "A human readable name",
            "description": "A detailed description about the asset",
            "version": "0.0.1",
            "distribution": "Ubuntu",
            "release": "14.04"
        }
        asset_dir = self.write_asset_directory("set_tarball_property", json.dumps(asset_spec))
        asset_bundle_id = json.loads(run('dx build_asset --json ' + asset_dir))['id']
        self.assertIn('record', asset_bundle_id)
        tarball_file_id = dxpy.describe(asset_bundle_id,
                                        fields={"details"})["details"]["archiveFileId"]["$dnanexus_link"]
        self.assertEqual(dxpy.describe(tarball_file_id,
                                       fields={"properties"})["properties"]["AssetBundle"], asset_bundle_id)

if __name__ == '__main__':
    if dxpy.AUTH_HELPER is None:
        sys.exit(1, 'Error: Need to be logged in to run these tests')
    if 'DXTEST_FULL' not in os.environ:
        if 'DXTEST_RUN_JOBS' not in os.environ:
            sys.stderr.write('WARNING: neither env var DXTEST_FULL nor DXTEST_RUN_JOBS are set; \
            tests that run jobs will not be run\n')
    unittest.main()



from __future__ import print_function


class JSONValidator(object):
    """
    JSON validator class to validate a JSON against a schema.

    The schema is a dictionary with the following structure:

    {
        "key1": {
            "type": dict,
            "properties": {
                "name": {"type": str, "required": True},
                "attribute": {"type": str}
            }
        },
        "key2": {
            "type": dict,
            "properties": {
                "xyz": {"type": str},
                "abc": {"type": list}
            },
            "conflicting_keys": [["xyz", "abc"]]
        },
        "list_key_example": {
            "type": list,
            "items": {
                "type": dict,
                "properties": {
                    "nested_key1": {"type": str, "required": True},
                    "nested_key2": {"type": str}
                }
            }
        },
        "id": {
            "type": list,
        },
        "conflicting_keys": [["key1", "key2"]],
        "dependent_conditional_keys": {
            "list_key_example": ["key1", "key2"]
        }
    }

    Key types can be defined via the "type" key

    Required keys can be defined via the "required" key

    Conflicting keys can be defined via the "conflicting_keys" key, where the value is a list of lists of keys that are mutually exclusive.
    Conflicting keys can be defined at the level of the schema for the entire JSON or at key-level.
    In the example above, within "key2" the keys "xyz" and "abc" are mutually exclusive and cannot be present together.
    Similarly, at the level of the whole JSON, "key1" and "key2" are mutually exclusive and cannot be present together.

    Conditional key combinations can be defined via the dependent_conditional_keys at the level of the schema. These should be defined as a dict where the value is a list.
    In the example above, dependent_conditional_keys is defined as {"list_key_example": ["key1", "key2"]}, which means whenever list_key_example is present, either "key1"
    or "key2" has to be also present.

    Currently the maximum level of nestedness supported for validation is as deep as shown in the example above via the "list_key_example" key.

    This is an example of a valid JSON that satisfies all the conditions defined in the schema above:

    {
        "key1": {
            "name": "test",
        },
        "list_key_example": [
            {"nested_key1": "1", "nested_key2": "2"},
            {"nested_key1": "3", "nested_key2": "4"},
            {"nested_key1": "5"}
        ],
        "id": ["id_1", "id_2"]
    }

    """

    def __init__(self, schema, error_handler=print):
        self.schema = schema
        self.error_handler = error_handler

        if not isinstance(schema, dict) or not schema:
            error_handler("Schema must be a non-empty dict.")

    def validate(self, input_json):
        if not isinstance(input_json, dict) or not input_json:
            self.error_handler("Input JSON must be a non-empty dict.")

        self.error_on_invalid_keys(input_json)

        self.validate_types(input_json)

        for key, value in self.schema.items():
            if key in input_json:
                self.validate_properties(value.get("properties", {}), input_json[key])

                if value.get("type") == list:
                    self.validate_list_items(
                        value.get("items", {}), input_json[key], key
                    )

                # Check for incompatible/conflicting subkeys defined at the key-level
                if value.get("conflicting_keys"):
                    self.check_incompatible_subkeys(input_json, key)
                    self.check_incompatible_subkeys(input_json[key], key)


        self.check_incompatible_keys(input_json)
        self.check_dependent_key_combinations(input_json)

    def validate_types(self, input_dict):
        for key, value in input_dict.items():
            if not isinstance(value, self.schema.get(key, {}).get("type")):
                self.error_handler(
                    "Key '{}' has an invalid type. Expected {} but got {}".format(
                        key, self.schema.get(key, {}).get("type"), type(value)
                    )
                )

    def validate_properties(self, properties, input_dict):
        for key, value in properties.items():
            if key not in input_dict and value.get("required"):
                self.error_handler(
                    "Required key '{}' was not found in the input JSON.".format(key)
                )
            if key in input_dict and not isinstance(input_dict[key], value.get("type")):
                self.error_handler(
                    "Key '{}' has an invalid type. Expected {} but got {}".format(
                        key, value.get("type"), type(input_dict[key])
                    )
                )

    def validate_list_items(self, item_schema, input_list, key_name):
        item_type = item_schema.get("type")
        if item_type:
            for item in input_list:
                if not isinstance(item, item_type):
                    self.error_handler(
                        "Expected list items within '{}' to be of type {} but got {} instead.".format(
                            key_name, item_type, type(item)
                        )
                    )
                self.validate_properties(item_schema.get("properties", {}), item)
        else:
            if not isinstance(input_list, list):
                self.error_handler(
                    "Expected list but got {} for {}".format(type(input_list), key_name)
                )
            for item in input_list:
                if not isinstance(item, str):
                    self.error_handler(
                        "Expected list items to be of type string for {}.".format(
                            key_name
                        )
                    )

    def check_incompatible_subkeys(self, input_json, current_key):
        for keys in self.schema.get(current_key, {}).get("conflicting_keys", []):
            if all(k in input_json for k in keys):
                self.error_handler(
                    "For {}, exactly one of {} must be provided in the supplied JSON object.".format(
                        current_key, " or ".join(keys)
                    )
                )

    def check_incompatible_keys(self, input_json):
        for keys in self.schema.get("conflicting_keys", []):
            if all(key in input_json for key in keys):
                self.error_handler(
                    "Exactly one of {} must be provided in the supplied JSON object.".format(
                        " or ".join(keys)
                    )
                )

    def check_dependent_key_combinations(
        self, input_json, enforce_one_associated_key=False
    ):
        mandatory_combinations = self.schema.get("dependent_conditional_keys", {})
        for main_key, associated_keys in mandatory_combinations.items():
            if main_key in input_json:
                present_associated_keys = [
                    key for key in associated_keys if key in input_json
                ]
                if len(present_associated_keys) == 0:
                    self.error_handler(
                        "When {} is present, one of the following keys must be also present: {}.".format(
                            main_key, ", ".join(associated_keys)
                        )
                    )
                if len(present_associated_keys) > 1 and enforce_one_associated_key:
                    self.error_handler(
                        "Only one of the associated keys {} can be present for main key {}".format(
                            ", ".join(associated_keys)
                        )
                    )

    def error_on_invalid_keys(self, input_json):
        invalid_keys = []
        for key in input_json:
            if key not in self.schema:
                invalid_keys.append(key)

        if invalid_keys:
            self.error_handler(
                "Found following invalid filters: {}".format(invalid_keys)
            )

    def are_list_items_within_range(
        self,
        input_json,
        key,
        start_subkey,
        end_subkey,
        window_width=250000000,
        check_each_separately=False,
    ):
        """
        This won't run by default when calling .validate() on a JSONValidator object. Run it separately when necessary
        """

        for item in input_json[key]:
            if int(item[end_subkey]) <= int(item[start_subkey]):
                self.error_handler(
                    "{} cannot be less than or equal to the {}".format(
                        end_subkey, start_subkey
                    )
                )

            if check_each_separately:
                if item[end_subkey] - item[start_subkey] > window_width:
                    self.error_handler(
                        "Range of {} and {} cannot be greater than {} for each item in {}".format(
                            start_subkey, end_subkey, window_width, key
                        )
                    )
        if not check_each_separately:
            # Check will be done across all items in the list
            # the entire search would be limited to window_width
            start_values = [int(item[start_subkey]) for item in input_json[key]]
            end_values = [int(item[end_subkey]) for item in input_json[key]]

            if max(end_values) - min(start_values) > window_width:
                self.error_handler(
                    "Range cannot be greater than {} for {}".format(window_width, key)
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2016 DNAnexus, Inc.
#
# This file is part of dx-toolkit (DNAnexus platform client libraries).
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may not
#   use this file except in compliance with the License. You may obtain a copy
#   of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

import os, sys, unittest, json, tempfile, subprocess
import pexpect

from dxpy_testutil import DXTestCase, check_output
import dxpy_testutil as testutil

import dxpy
from dxpy.utils.completer import InstanceTypesCompleter
import pytest

supported_languages = ['Python', 'bash']

spawn_extra_args = {"encoding" : "utf-8" }

def run_dx_app_wizard(instance_type=None):
    old_cwd = os.getcwd()
    tempdir = tempfile.mkdtemp(prefix='Программа')
    os.chdir(tempdir)
    try:
        wizard = pexpect.spawn("dx-app-wizard --template parallelized",
                               **spawn_extra_args)
        wizard.logfile = sys.stdout
        wizard.setwinsize(20, 90)
        wizard.expect("App Name:")
        wizard.sendline("Имя")
        wizard.expect("The name of your app must match")
        wizard.expect("App Name:")
        wizard.sendline("MyTestApp")
        wizard.expect("Title")
        wizard.sendline("Заголовок")
        wizard.expect("Summary")
        wizard.sendline("Конспект")
        wizard.expect("Version")
        wizard.sendline("1.2.3")
        wizard.expect("1st input name")
        wizard.sendline("in1")
        wizard.expect("Label")
        wizard.sendline("Метка")
        wizard.expect("Choose a class")
        wizard.sendline("int")
        wizard.expect("optional parameter")
        wizard.sendline("n")
        wizard.expect("2nd input name")
        wizard.sendline()
        wizard.expect("1st output name")
        wizard.sendline("out1")
        wizard.expect("Label")
        wizard.sendline()
        wizard.expect("Choose a class")
        wizard.sendline("целое")
        wizard.expect("Not a recognized class")
        wizard.sendline("int")
        wizard.expect("2nd output name")
        wizard.sendline()
        wizard.expect("Timeout policy")
        wizard.sendline("31d")
        wizard.expect("Error: max allowed timeout is 30 days")
        wizard.sendline("ЄЯTЪЦGЇCЄкЇ")
        wizard.expect("Error: enter an int with a single-letter suffix")
        wizard.expect("Timeout policy")
        wizard.sendline("24h")
        wizard.expect("Programming language")
        wizard.sendline("АЛГОЛ")
        wizard.expect("Error: unrecognized response")
        wizard.sendline("Python")
        wizard.expect("Will this app need access to the Internet?")
        wizard.sendline("y")
        wizard.expect("Will this app need access to the parent project?")
        wizard.sendline("y")
        wizard.expect("Choose an instance type for your app")
        wizard.sendline("t1.микро")
        wizard.expect("Error: unrecognized response, expected one of")
        wizard.expect("Choose an instance type for your app")
        if instance_type is not None:
            wizard.sendline(instance_type)
        else:
            wizard.sendline()
        wizard.expect("App directory created")
        wizard.close()

        appdir = os.path.join(tempdir, "MyTestApp")
        return appdir
    finally:
        os.chdir(old_cwd)

def create_app_dir_with_dxapp_json(dxapp_json, language):
    old_cwd = os.getcwd()
    tempdir = tempfile.mkdtemp()
    os.chdir(tempdir)
    try:
        with open('dxapp.json', 'w') as fd:
            json.dump(dxapp_json, fd)

        wizard = pexpect.spawn("dx-app-wizard --json-file dxapp.json --language " + language,
                               **spawn_extra_args)
        wizard.setwinsize(20, 90)
        wizard.logfile = sys.stdout
        wizard.expect("App Name")
        wizard.sendline()
        wizard.expect("Version")
        wizard.sendline()
        wizard.expect("Timeout policy")
        wizard.sendline()
        wizard.expect("Will this app need access to the Internet?")
        wizard.sendline()
        wizard.expect("Will this app need access to the parent project?")
        wizard.sendline()
        wizard.expect("Choose an instance type for your app")
        wizard.sendline()
        wizard.expect("App directory created")
        wizard.close()

        appdir = os.path.join(tempdir, dxapp_json['name'])
        return appdir
    finally:
        os.chdir(old_cwd)

class TestDXAppWizard(DXTestCase):
    def test_invalid_arguments(self):
        with self.assertRaises(testutil.DXCalledProcessError):
            check_output(['dx-app-wizard', '--template=par'])

    @pytest.mark.TRACEABILITY_MATRIX
    @testutil.update_traceability_matrix(["DNA_CLI_HELP_CREATE_APP_WIZARD"])
    def test_dx_app_wizard(self):
        appdir = run_dx_app_wizard()
        with open(os.path.join(appdir, 'dxapp.json')) as fh:
            dxapp_json = json.load(fh)
        self.assertEqual(dxapp_json['regionalOptions']['aws:us-east-1']['systemRequirements']['*']['instanceType'],
                         InstanceTypesCompleter.default_instance_type.Name)
        self.assertEqual(dxapp_json['runSpec']['distribution'], 'Ubuntu')
        self.assertEqual(dxapp_json['runSpec']['release'], '24.04')
        self.assertEqual(dxapp_json['runSpec']['version'], '0')
        self.assertEqual(dxapp_json['runSpec']['interpreter'], 'python3')
        self.assertEqual(dxapp_json['runSpec']['timeoutPolicy']['*']['hours'], 24)

    def test_dx_app_wizard_with_azure_instance_type(self):
        appdir = run_dx_app_wizard("azure:mem1_ssd1_x2")
        with open(os.path.join(appdir, 'dxapp.json')) as fh:
            dxapp_json = json.load(fh)
        self.assertEqual(dxapp_json['regionalOptions']['azure:westus']['systemRequirements']['*']['instanceType'],
                         "azure:mem1_ssd1_x2")

    def test_var_initialization(self):
        '''
        This test assumes a well-formed input spec and mostly just
        tests that everything compiles and the variable initialization
        code does not throw any errors.
        '''

        print("Setting current project to", self.project)
        dxpy.WORKSPACE_ID = self.project
        dxpy.PROJECT_CONTEXT_ID = self.project

        # Make some data objects for input
        dxpy.api.applet_new({"project": dxpy.WORKSPACE_ID,
                             "name": "anapplet",
                             "dxapi": "1.0.0",
                             "runSpec": {"code": "", "interpreter": "bash",
                                         "distribution": "Ubuntu", "release": "14.04"}})['id']
        dxpy.upload_string("foo", name="afile")
        dxrecord = dxpy.new_dxrecord(name="arecord")
        dxrecord.close()

        dxapp_json = {
            "name": "all_vars",
            "title": "all_vars",
            "summary": "all_vars",
            "dxapi": "1.0.0",
            "version": "0.0.1",
            "categories": [],
            "inputSpec": [],
            "outputSpec": []
        }

        classes = ['applet', 'record', 'file',
                   'boolean', 'int', 'float', 'string', 'hash',
                   'array:applet', 'array:record', 'array:file',
                   'array:boolean', 'array:int', 'array:float', 'array:string']

        for classname in classes:
            dxapp_json['inputSpec'].append({"name": "required_" + classname.replace(":", "_"),
                                            "class": classname,
                                            "optional": False})
            # Note: marking outputs as optional so that empty arrays
            # will be acceptable; keeping names the same (as required)
            # in order to allow pass-through from input variables
            dxapp_json['outputSpec'].append({"name": "required_" + classname.replace(":", "_"),
                                             "class": classname,
                                             "optional": True})
            dxapp_json['inputSpec'].append({"name": "optional_" + classname.replace(":", "_"),
                                            "class": classname,
                                            "optional": True})

        cmdline_args = ['-irequired_applet=anapplet',
                        '-irequired_array_applet=anapplet',
                        '-irequired_record=arecord',
                        '-irequired_array_record=arecord',
                        '-irequired_file=afile',
                        '-irequired_array_file=afile',
                        '-irequired_boolean=true',
                        '-irequired_array_boolean=true',
                        '-irequired_array_boolean=false',
                        '-irequired_int=32',
                        '-irequired_array_int=42',
                        '-irequired_float=3.4',
                        '-irequired_array_float=.42',
                        '-irequired_string=foo',
                        '-irequired_array_string=bar',
                        '-irequired_hash={"foo":"bar"}']
        for lang in supported_languages:
            appdir = create_app_dir_with_dxapp_json(dxapp_json, lang)

            if testutil.TEST_RUN_JOBS:
                # Now actually make it an applet and run it
                applet_name = dxapp_json['name'] + '-' + lang
                subprocess.check_output(['dx', 'build', appdir, '--destination', applet_name])
                subprocess.check_output(['dx', 'run', applet_name, '-y', '--wait'] + cmdline_args)


if __name__ == '__main__':
    unittest.main()
